from llvmlite import binding
from numba import types
import pathlib
import platform

# interface to the c implementation for use in numba

library_path = pathlib.Path(__file__).parent.absolute()

if platform.system() == "Windows":
    library_name = "_ckdtree.lib"
else:
    library_name = "_ckdtree.so"


# a simple class holding all the function pointers required from the c library
class CKDTreeFunctions(object):
    pass


ckdtree = CKDTreeFunctions

# define the pointer types to be used
float_ptr = types.CPointer(types.float32)
double_ptr = types.CPointer(types.double)
ssize_t_ptr = types.CPointer(types.ssize_t)
ckdtree_handle = types.size_t

try:
    binding.load_library_permanently(str(library_path / library_name))
except Exception:
    raise ImportError('Unable to load dynamic library.')


# the data types supported by the c implementation
_supported_types = {types.float32: ("float", float_ptr),
                    types.float64: ("double", double_ptr)
                    }

ckdtree.init = {}
ckdtree.free = {}
ckdtree.build = {}
ckdtree.size = {}
ckdtree.query_knn = {}
ckdtree.query_radius = {}
ckdtree.leafsize = {}
ckdtree.nodesize = {}
ckdtree.copy_tree = {}


for numba_type, (c_type, ptr_type) in _supported_types.items():
    ckdtree.init[numba_type] = types.ExternalFunction(f"ckdtree_init_{c_type}", ckdtree_handle(
        types.voidptr,
        types.ssize_t,
        ptr_type,
        ssize_t_ptr,
        types.ssize_t,
        types.ssize_t,
        types.ssize_t,
        ptr_type,
        ptr_type
    ))

    ckdtree.free[numba_type] = types.ExternalFunction(f"ckdtree_free_{c_type}", types.void(
        ckdtree_handle
    ))

    ckdtree.build[numba_type] = types.ExternalFunction(f"ckdtree_build_{c_type}", types.int32(
        ckdtree_handle,
        types.ssize_t,
        types.ssize_t,
        ptr_type,
        ptr_type,
        types.int32,
        types.int32
    ))

    ckdtree.size[numba_type] = types.ExternalFunction(f"ckdtree_size_{c_type}", types.ssize_t(
        ckdtree_handle
    ))

    ckdtree.query_knn[numba_type] = types.ExternalFunction(f"ckdtree_query_knn_{c_type}", types.int32(
        ckdtree_handle,
        ptr_type,
        ssize_t_ptr,
        ssize_t_ptr,
        ptr_type,
        types.ssize_t,
        types.ssize_t,
        types.double,
        types.double,
        numba_type
    ))

    ckdtree.query_radius[numba_type] = types.ExternalFunction(f"ckdtree_query_radius_{c_type}", types.voidptr(
        ckdtree_handle,
        ptr_type,
        types.ssize_t,
        numba_type,
        types.double,
        types.double,
        types.bool_,
        types.bool_
    ))

    ckdtree.leafsize[numba_type] = types.ExternalFunction(f"leafsize_{c_type}", types.ssize_t(
        ckdtree_handle
    ))

    ckdtree.nodesize[numba_type] = types.ExternalFunction(f"nodesize_{c_type}", types.ssize_t(
        ckdtree_handle
    ))

    ckdtree.copy_tree[numba_type] = types.ExternalFunction(f"copy_tree_{c_type}", types.ssize_t(
        ckdtree_handle,
        types.voidptr
    ))


ckdtree.radius_result_set_get_size = types.ExternalFunction("radius_result_set_get_size", types.ssize_t(
    types.voidptr
))

ckdtree.radius_result_set_copy_and_free = types.ExternalFunction("radius_result_set_copy_and_free", types.void(
    types.voidptr,
    ssize_t_ptr
))       




import pytest
from numba_kdtree import KDTree, KDTreeType
import numpy as np
from scipy.spatial import cKDTree
import numba as nb
from timeit import timeit


@pytest.fixture(scope='module')
def data():
    from .util import load_standford_bunny
    return load_standford_bunny(dtype=np.float32)


@pytest.fixture(scope='module')
def kdtree(data) -> KDTreeType:
    kd_tree = KDTree(data, leafsize=10, balanced=False, compact=False)
    return kd_tree


@pytest.fixture(scope='module')
def scipy_kdtree(data):
    scipy_tree = cKDTree(data, leafsize=10, balanced_tree=False, compact_nodes=False)
    return scipy_tree


def _test_kdtree_build(data, **kwargs):
    kd_tree = KDTree(data, **kwargs)
    assert kd_tree.built()
    return kd_tree


def test_kdtree_build(data):
    kdtree_f = _test_kdtree_build(data, leafsize=10, balanced=False, compact=False)

    assert kdtree_f.size == 13143
    data_double = data.astype(np.float64)
    kdtree_d = _test_kdtree_build(data_double, leafsize=10, balanced=False, compact=False)

    scipy_tree = cKDTree(data, leafsize=10, balanced_tree=False, compact_nodes=False)
    assert kdtree_f.size == kdtree_d.size == scipy_tree.size, "invalid_sizes kdtree"

    num_executions = 5
    runtime_kdtree_f = timeit(lambda: _test_kdtree_build(data,
                                                         leafsize=10, balanced=False, compact=False),
                              number=num_executions) / num_executions

    runtime_kdtree_d = timeit(lambda: _test_kdtree_build(data_double,
                                                         leafsize=10, balanced=False, compact=False),
                              number=num_executions) / num_executions
    runtime_scipy = timeit(lambda: cKDTree(data,
                                           leafsize=10, balanced_tree=False, compact_nodes=False),
                           number=num_executions) / num_executions

    print("\nbuild time:\nscipy:", runtime_scipy, "\nkdtree(float):",
          runtime_kdtree_f, "\nkdtree(double):", runtime_kdtree_d)


def test_kdtree_query(data, kdtree: KDTreeType, scipy_kdtree):
    k = 10
    dd, ii, nn = kdtree.query(data[:1], k=k)  # pre-compile

    # query the nearest neighbors of each input point in a single thread
    dd, ii, nn = kdtree.query(data, k=k)
    dd_scipy, ii_scipy = scipy_kdtree.query(data, k=k, workers=1)

    assert np.all(ii == ii_scipy)
    assert np.allclose(dd, dd_scipy)

    num_executions = 5
    k_benchmark = 30
    runtime_kdtree_query = timeit(lambda: kdtree.query(data, k=k_benchmark),
                              number=num_executions) / num_executions
    runtime_scipy_query = timeit(lambda: scipy_kdtree.query(data, k=k_benchmark, workers=1),
                              number=num_executions) / num_executions
    print("\nquery time(single thread):\nscipy:", runtime_scipy_query, "\nkdtree(float):",
          runtime_kdtree_query)


def test_kdtree_query_parallel(data, kdtree, scipy_kdtree):
    k = 10
    #dd, ii, nn = kdtree.query_parallel(data[:1], k=k)  # pre-compile
    
    # query using all available cpu cores
    dd, ii, nn = kdtree.query_parallel(data, k=k)  # pre-compile
    dd_scipy, ii_scipy = scipy_kdtree.query(data, k=k, workers=-1)

    assert np.all(ii == ii_scipy)
    assert np.allclose(dd, dd_scipy)

    num_executions = 5
    k_benchmark = 30
    runtime_kdtree_query = timeit(lambda: kdtree.query_parallel(data, k=k_benchmark),
                              number=num_executions) / num_executions
    runtime_scipy_query = timeit(lambda: scipy_kdtree.query(data, k=k_benchmark, workers=-1),
                              number=num_executions) / num_executions
    print("\nquery time(multi thread):\nscipy:", runtime_scipy_query, "\nkdtree(float):",
          runtime_kdtree_query)


def test_kdtree_query_max_radius(data, kdtree, scipy_kdtree):
    k = 100
    upper_bound = 0.002

    dd, ii, nn = kdtree.query(data[:1], k=k, distance_upper_bound=upper_bound)  # pre-compile

    # query the nearest neighbors of each input point in a single thread
    dd, ii, nn = kdtree.query(data, k=k, distance_upper_bound=upper_bound)
    dd_scipy, ii_scipy = scipy_kdtree.query(data, k=k, distance_upper_bound=upper_bound, workers=1)
    assert np.any(nn != k)

    invalid = ii < 0
    assert np.all(ii_scipy[invalid] == data.shape[0])
    assert np.all(ii[~invalid] == ii_scipy[~invalid])

    num_executions = 5
    k_benchmark = 100
    runtime_kdtree_query = timeit(lambda: kdtree.query(data, k=k_benchmark, distance_upper_bound=upper_bound),
                              number=num_executions) / num_executions
    runtime_scipy_query = timeit(lambda: scipy_kdtree.query(data, k=k_benchmark, workers=1,
                                                            distance_upper_bound=upper_bound),
                              number=num_executions) / num_executions
    print("\nquery time(single threadm, max radius):\nscipy:", runtime_scipy_query, "\nkdtree(float):",
          runtime_kdtree_query)


def test_kdtree_query_radius(data, kdtree, scipy_kdtree):
    # pre compile
    ii = kdtree.query_radius(data[:1], r=0.01, return_sorted=True)

    ii = kdtree.query_radius(data[:100], r=0.01, return_sorted=True)
    ii_scipy = scipy_kdtree.query_ball_point(data[:100], r=0.01, return_sorted=True, workers=1)

    for i in range(len(ii)):
        assert len(ii[i]) == len(ii_scipy[i])
        assert np.all(ii[i] == ii_scipy[i]), "Not equal for i={}".format(i)

    num_executions = 5
    r_benchmark = 0.1
    runtime_kdtree_query = timeit(lambda: kdtree.query_radius(data[:500], r=r_benchmark, return_sorted=True),
                              number=num_executions) / num_executions
    runtime_scipy_query = timeit(lambda: scipy_kdtree.query_ball_point(data[:500], r=r_benchmark,
                                                                       return_sorted=True, workers=1),
                              number=num_executions) / num_executions
    print("\nradius query time(single thread):\nscipy:", runtime_scipy_query, "\nkdtree(float):",
          runtime_kdtree_query)


def test_kdtree_query_radius_parallel(data, kdtree, scipy_kdtree):

    ii = kdtree.query_radius_parallel(data[:100], r=0.01, return_sorted=True)
    ii_scipy = scipy_kdtree.query_ball_point(data[:100], r=0.01, return_sorted=True, workers=-1)

    for i in range(len(ii)):
        assert len(ii[i]) == len(ii_scipy[i])
        assert np.all(ii[i] == ii_scipy[i]), "Not equal for i={}".format(i)

    num_executions = 5
    r_benchmark = 0.1
    runtime_kdtree_query = timeit(lambda: kdtree.query_radius_parallel(data[:500], r=r_benchmark, return_sorted=True),
                              number=num_executions) / num_executions
    runtime_scipy_query = timeit(lambda: scipy_kdtree.query_ball_point(data[:500], r=r_benchmark,
                                                                       return_sorted=True, workers=-1),
                              number=num_executions) / num_executions
    print("\nradius query time(multi thread):\nscipy:", runtime_scipy_query, "\nkdtree(float):",
          runtime_kdtree_query)
    

def test_kdtree_query_radius_array(data, kdtree, scipy_kdtree):
    # pre compile
    ii = kdtree.query_radius(data[:1], r=0.01, return_sorted=True)

    radii = np.linspace(0.01, 0.05, 100)
    ii = kdtree.query_radius(data[:100], r=radii, return_sorted=True)
    ii_scipy = scipy_kdtree.query_ball_point(data[:100], r=radii, return_sorted=True, workers=1)

    for i in range(len(ii)):
        assert len(ii[i]) == len(ii_scipy[i])
        assert np.all(ii[i] == ii_scipy[i]), "Not equal for i={}".format(i)

    # invalid array shapes should not work
    with pytest.raises(ValueError):
        ii_invalid = kdtree.query_radius(data[:100], r=radii[:50], return_sorted=True)

    num_executions = 5
    r_benchmark = np.linspace(0.01, 0.05, 500)
    runtime_kdtree_query = timeit(lambda: kdtree.query_radius(data[:500], r=r_benchmark, return_sorted=True),
                              number=num_executions) / num_executions
    runtime_scipy_query = timeit(lambda: scipy_kdtree.query_ball_point(data[:500], r=r_benchmark,
                                                                       return_sorted=True, workers=1),
                              number=num_executions) / num_executions
    print("\nradius query time(single thread):\nscipy:", runtime_scipy_query, "\nkdtree(float):",
          runtime_kdtree_query)
    

def test_kdtree_query_radius_conversion(data, kdtree):
    # pre compile
    radius = [0.01]
    ii = kdtree.query_radius(data[:1], r=radius, return_sorted=True)

    radii = np.linspace(0.01, 0.05, 100).tolist()
    ii = kdtree.query_radius(data[:100], r=radii, return_sorted=True)
    ii_test = kdtree.query_radius(data[:100], r=np.asarray(radii), return_sorted=True)

    for i in range(len(ii)):
        assert len(ii[i]) == len(ii_test[i])
        assert np.all(ii[i] == ii_test[i]), "Not equal for i={}".format(i)

    # invalid array shapes should not work
    with pytest.raises(ValueError):
        ii_invalid = kdtree.query_radius(data[:100], r=radii[:50], return_sorted=True)

    num_executions = 5
    r_benchmark = np.linspace(0.01, 0.05, 500).tolist()
    runtime_kdtree_query = timeit(lambda: kdtree.query_radius(data[:500], r=r_benchmark, return_sorted=True),
                              number=num_executions) / num_executions
    print("\nradius query time(single thread): kdtree(float):",
          runtime_kdtree_query)


def test_kdtree_query_radius_array_parallel(data, kdtree, scipy_kdtree):
    radii = np.linspace(0.01, 0.05, 100)
    ii = kdtree.query_radius_parallel(data[:100], r=radii, return_sorted=True)
    ii_scipy = scipy_kdtree.query_ball_point(data[:100], r=radii, return_sorted=True, workers=-1)

    for i in range(len(ii)):
        assert len(ii[i]) == len(ii_scipy[i])
        assert np.all(ii[i] == ii_scipy[i]), "Not equal for i={}".format(i)

    num_executions = 5
    r_benchmark = np.linspace(0.01, 0.05, 500)
    runtime_kdtree_query = timeit(lambda: kdtree.query_radius_parallel(data[:500], r=r_benchmark, return_sorted=True),
                              number=num_executions) / num_executions
    runtime_scipy_query = timeit(lambda: scipy_kdtree.query_ball_point(data[:500], r=r_benchmark,
                                                                       return_sorted=True, workers=-1),
                              number=num_executions) / num_executions
    print("\nradius query time(single thread):\nscipy:", runtime_scipy_query, "\nkdtree(float):",
          runtime_kdtree_query)


def test_argument_conversion(data, kdtree):
    import warnings
    # conversion of scalar arguments
    _, ii_test, _ = kdtree.query(data[0], k=10)
    with warnings.catch_warnings():
        # conversion of lists (this is deprecated but should still work as long as the typed list is not the default yet)
        warnings.simplefilter("ignore", nb.errors.NumbaPendingDeprecationWarning)
        _, ii, _ = kdtree.query([data[0]], k=10)
        assert np.all(ii == ii_test)

    # conversion of numba lists
    _, ii, _ = kdtree.query(nb.typed.List([data[0]]), k=10)
    assert np.all(ii == ii_test)

    # conversion of manually specified arrays
    _, ii_test, _ = kdtree.query(np.array([0, 0, 0]), k=10)
    with warnings.catch_warnings():
        # conversion of lists (this is deprecated but should still work as long as the typed list is not the default yet)
        warnings.simplefilter("ignore", nb.errors.NumbaPendingDeprecationWarning)
        _, ii2, _ = kdtree.query([0, 0, 0], k=10)
    assert np.all(ii2 == ii_test)


def test_use_in_numba_function(data, kdtree):
    @nb.njit(nogil=True, fastmath=True)
    def query_numba(data, kdtree, k):
        ii = np.empty((data.shape[0], k), dtype=np.int64)
        for i in range(data.shape[0]):
            d, ii[i], n = kdtree.query(data[i], k=k)

        return ii

    # pre compile
    ii = query_numba(data, kdtree, 10)
    dd, ii, nn = kdtree.query(data[:1], k=10)  # pre-compile

    num_executions = 5
    k_benchmark = 30

    # run times should be very similar here
    runtime_kdtree_query = timeit(lambda: kdtree.query(data, k=k_benchmark),
                              number=num_executions) / num_executions
    runtime_numba_kdtree_query = timeit(lambda: query_numba(data, kdtree, k_benchmark),
                              number=num_executions) / num_executions
    print("\nnumba query time(single thread):\nkdtree(float):",
          runtime_kdtree_query, "\nkdtree (numba):", runtime_numba_kdtree_query)


def test_use_in_numba_function_parallel(data, kdtree):
    @nb.njit(nogil=True, fastmath=True, cache=True, parallel=True)
    def query_numba_parallel(data, kdtree, k):
        ii = np.empty((data.shape[0], k), dtype=np.int64)
        for i in nb.prange(data.shape[0]):
            d, ii[i], n = kdtree.query(data[i], k=k)

        return ii

    # pre compile
    ii = query_numba_parallel(data, kdtree, 10)
    dd, ii, nn = kdtree.query_parallel(data[:1], k=10)  # pre-compile

    num_executions = 5
    k_benchmark = 30

    # run times should be very similar here
    runtime_kdtree_query = timeit(lambda: kdtree.query_parallel(data, k=k_benchmark),
                              number=num_executions) / num_executions
    runtime_numba_kdtree_query = timeit(lambda: query_numba_parallel(data, kdtree, k_benchmark),
                              number=num_executions) / num_executions
    print("\nnumba query time(multi thread):\nkdtree(float):",
          runtime_kdtree_query, "\nkdtree (numba):", runtime_numba_kdtree_query)


def test_construct_in_numba_function(data):
    @nb.njit(nogil=True, fastmath=True)
    def construct_kdtree_in_numba(data, compact=False, balanced=False):
        kdtree = KDTree(data, compact=compact, balanced=balanced)
        return kdtree

    num_executions = 10
    runtime_kdtree_numba = timeit(lambda: construct_kdtree_in_numba(data, compact=True, balanced=True),
                              number=num_executions) / num_executions

    runtime_kdtree_python = timeit(lambda: KDTree(data), number=num_executions) / num_executions

    print("\nbuild time:\nnumba:",
          runtime_kdtree_numba, "\npython:", runtime_kdtree_python)


def test_pickle(data, kdtree):
    import pickle

    # pass the tree through pickle
    serialized_tree = pickle.dumps(kdtree)

    restored_kdtree = pickle.loads(serialized_tree)

    k = 10
    # query the old tree
    dd, ii, nn = kdtree.query_parallel(data[:100], k=k)

    # query the new tree
    dd_r, ii_r, nn_r = restored_kdtree.query_parallel(data[:100], k=k)

    assert np.allclose(dd, dd_r)
    assert np.all(ii == ii_r)
    assert np.all(nn == nn_r)


try:
    from importlib.metadata import version

    __version__ = version("numba_kdtree")
except Exception:  # pragma: no cover # pylint: disable=broad-exception-caught
    try:
        from ._version import __version__
    except ImportError:
        __version__ = '0.0.0'

from .kd_tree import *



from __future__ import annotations
import numba as nb
import numpy as np

from numba.core import types, cgutils
from numba.experimental import structref
from numba.extending import overload_method, intrinsic
from .ckdtree import ckdtree as ckdtree_ct
import warnings
from typing import Optional, Any


__all__ = ["KDTree", "KDTreeType"]

INT_TYPE = np.int64
INT_TYPE_T = nb.int64

DIM_TYPE = np.uint32
DIM_TYPE_T = nb.uint32

FLOAT_TYPE = np.float32
FLOAT_TYPE_T = nb.float32

BOOL_TYPE = np.uint8
BOOL_TYPE_T = nb.uint8

IntArray = np.ndarray
DimIndexArray = np.ndarray
DataArray = np.ndarray
BoolArray = np.ndarray

NUMBA_THREADS = nb.config.NUMBA_NUM_THREADS


@nb.njit(nogil=True, inline='always', cache=True)
def _list_to_2d_array(arraylist, dtype):
    n = len(arraylist)
    k = arraylist[0].shape[0]
    array = np.zeros((n, k), dtype)
    for i in range(n):
        array[i] = arraylist[i]
    return array


def _convert_to_valid_input(X, n_features, dtype):
    # this is a stub for numba overload
    pass


@nb.extending.overload(_convert_to_valid_input, jit_options={'nogil': True, 'fastmath': True, "cache": True})
def _ol_restore_kdtree_impl(X, n_features, dtype):
    convert_list_to_array = isinstance(X, (nb.types.ListType, nb.types.List)) and isinstance(X.dtype, nb.types.ArrayCompatible)

    def _convert_impl(X, n_features, dtype):
        if convert_list_to_array:
            x_tmp = _list_to_2d_array(X, dtype)
        else:
            x_tmp = np.asarray(X, dtype=dtype)
        return np.ascontiguousarray(x_tmp).reshape(-1, n_features)

    return _convert_impl


@structref.register
class KDTreeNumbaType(types.StructRef):
    def preprocess_fields(self, fields):
        # We don't want the struct to take Literal types.
        return tuple((name, types.unliteral(typ)) for name, typ in fields)


class KDTreeType(structref.StructRefProxy):
    def __new__(cls, ckdtree, root_bbox, data, idx):
        return structref.StructRefProxy.__new__(cls,
                                                ckdtree,
                                                root_bbox,
                                                data,
                                                idx)
    @property
    def root_bbox(self) -> DataArray:
        """Returns the root bounding box of the kdtree.

        Returns:
            Numpy array of shape (2, n_features).
        """
        return _KDTree_get_root_bbox(self)

    @property
    def data(self) -> DataArray:
        """Returns the underlying data array.

        Returns:
            Numpy array of shape (n, n_features)
        """
        return _KDTree_get_data(self)

    @property
    def idx(self) -> DataArray:
        """Returns the underlying index array of all points in the data array sorted by the tree traversal order.

        Returns:
            Numpy array of shape (n,)
        """
        return _KDTree_get_idx(self)

    @property
    def size(self) -> int:
        """Returns the number of nodes in the kdtree. 
        """
        return _KDTree_get_size(self)
    
    @property
    def leafsize(self) -> int:
        """Returns the size of a leaf in the kdtree in bytes.
        """
        return _KDTree_get_leafsize(self)

    def built(self) -> bool:
        """Returns True, if the kdtree has been built already, False otherwise.
        """
        return _KDTree_built(self)

    def __del__(self) -> None:
        try:
            self.free_index()
        except (ModuleNotFoundError, ImportError):
            # HACK: we are in the process of shutting down the interpreter so calling the external c function
            # might not be possible any more. For now just ignore this
            pass

    def __reduce__(self) -> Any:
        """Pickle support
        """
        args = _KDTree_reduce_args(self)
        return _restore_kdtree, args

    def free_index(self) -> None:
        """free the internal index and ressources. The kdtree has to be rebuilt afterwards to be used.
        """
        _KDTree_free(self)

    def query(self, 
              X: DataArray, 
              k: int = 1, 
              p: float = 2.0, 
              eps: float = 0.0, 
              distance_upper_bound: float = np.inf) -> tuple[DataArray, DataArray, DataArray]:
        """Query the k nearest neighbors of the given query points. The results are returned as 3 arrays containing the 
        distances, indices, and number of found neighbors. As the distance and index arrays are allocated before the query, the number of
        neighbors found has to be checked against the returned number of found neighbors. 
        Indices and distances above the number of neighbors found are invalid.

        Args:
            X: The query points as an array of shape (n, n_features) or (n_features,).
            k: The number of neighbors to search. Defaults to 1.
            p: The distance metric (p-norm) to use. Defaults to 2.0 (euclidean norm).
            eps: Optional epsilon for approximative search. Defaults to 0.0 (exact search).
            distance_upper_bound: The upper bound of the distance between the query and the found neighbors. Defaults to np.inf.

        Returns:
            A tuple of:
                - Numpy array of shape (n, k) containing the distances between each query point and it's k nearest neighbors.
                - Numpy array of shape (n, k) containng the indices of the nearest neighbors for each query point.
                - Numpy array of shape (n,) containing the number of neighbors found for each query point. 
                  This is usually equal to k, except of the distance_upper_bound has been set.
        """
        return _KDTree_query(self, X, k, p, eps, distance_upper_bound)
    
    def query_parallel(self, 
              X: DataArray, 
              k: int = 1, 
              p: float = 2.0, 
              eps: float = 0.0, 
              distance_upper_bound: float = np.inf) -> tuple[DataArray, DataArray, DataArray]:
        """Query to k nearest neighbors of the given query points using a parallel implementation with one thread per query point. 
        See query() for details.
        """
        return _KDTree_query_parallel(self, X, k, p, eps, distance_upper_bound)

    def query_radius(self, X: DataArray, 
                     r: DataArray | float, 
                     p: float = 2.0, 
                     eps: float = 0.0, 
                     return_sorted: bool = False, 
                     return_length: bool = False) -> list[DataArray]:
        """Query all neighbors within a given radius for each query point.

        Args:
            X: The query points as an array of shape (n, n_features) or (n_features,).
            r: The search radius. If given as an array of shape (n,), a different radius will be used for each query point.
            p: The distance metric (p-norm) to use. Defaults to 2.0 (euclidean norm).
            eps: Optional epsilon for approximative search. Defaults to 0.0 (exact search).
            return_sorted: Return the neighbors sorted by distance. Defaults to False.
            return_length: Returns the number of neighbors found instead of their indices. Defaults to False.

        Returns:
            A list of numpy arrays containing either the indices of each neighbor or the number of neihbors found within the given search radius.
        """
        return _KDTree_query_radius(self, X, r, p, eps, return_sorted, return_length)
    
    def query_radius_parallel(self, X: DataArray, 
                     r: DataArray | float, 
                     p: float = 2.0, 
                     eps: float = 0.0, 
                     return_sorted: bool = False, 
                     return_length: bool = False) -> list[DataArray]:
        """Query all neighbors within a given radius for each query point using a parallel implementation with one thread per query point.
        See query_radius() for details.
        """
        return _KDTree_query_radius_parallel(self, X, r, p, eps, return_sorted, return_length)


structref.define_proxy(KDTreeType, KDTreeNumbaType,
                       ["ckdtree", "root_bbox", "data", "idx"])


# define wrapper functions for each method of the kdtree
@nb.njit(cache=True)
def _KDTree_get_root_bbox(self):
    return self.root_bbox


@nb.njit(cache=True)
def _KDTree_get_data(self):
    return self.data


@nb.njit(cache=True)
def _KDTree_get_idx(self):
    return self.idx


@nb.njit(cache=True)
def _KDTree_get_size(self):
    return self._size()


@nb.njit(cache=True)
def _KDTree_get_leafsize(self):
    return self._leafsize()


@nb.njit(cache=True)
def _KDTree_reduce_args(self):
    return self._reduce_args()


@nb.njit(cache=True)
def _KDTree_built(self):
    return self.ckdtree != 0


@nb.njit(cache=True)
def _KDTree_free(self):
    self.free_index()


@nb.njit(cache=True)
def _KDTree_query(self, X, k=1, p=2, eps=0.0, distance_upper_bound=np.inf):
    return self.query(X, k=k, p=p, eps=eps, distance_upper_bound=distance_upper_bound)


@nb.njit(cache=False)
def _KDTree_query_parallel(self, X, k=1, p=2, eps=0.0, distance_upper_bound=np.inf):
    return self.query_parallel(X, k=k, p=p, eps=eps, distance_upper_bound=distance_upper_bound)


@nb.njit(cache=True)
def _KDTree_query_radius(self, X, r, p=2.0, eps=0.0, return_sorted=False, return_length=False):
    return self.query_radius(X, r, p, eps, return_sorted, return_length)


@nb.njit(cache=False)
def _KDTree_query_radius_parallel(self, X, r, p=2.0, eps=0.0, return_sorted=False, return_length=False):
    return self.query_radius_parallel(X, r, p, eps, return_sorted, return_length)


# functions required for pickling the Kdtree
@overload_method(KDTreeNumbaType, "_size", jit_options={"cache": True})
def _ol_size(self):
    dtype = self.field_dict['data'].dtype
    if dtype != nb.types.float32:
        dtype = nb.types.float64

    func_size = ckdtree_ct.size[dtype]

    def _size_impl(self):
        return func_size(self.ckdtree)

    return _size_impl


@overload_method(KDTreeNumbaType, "_leafsize", jit_options={"cache": True})
def _ol_leafsize(self):
    """Returns the leaf size of the underlying tree
    """
    dtype = self.field_dict['data'].dtype
    if dtype != nb.types.float32:
        dtype = nb.types.float64

    func_leafsize = ckdtree_ct.leafsize[dtype]

    def _leafsize_impl(self):
        return func_leafsize(self.ckdtree)

    return _leafsize_impl


@overload_method(KDTreeNumbaType, "free_index", jit_options={"cache": True})
def _ol_free_index(self):
    dtype = self.field_dict['data'].dtype
    if dtype != nb.types.float32:
        dtype = nb.types.float64
    func_free = ckdtree_ct.free[dtype]

    def _free_index_impl(self):
        func_free(self.ckdtree)
        self.ckdtree = 0

    return _free_index_impl


@overload_method(KDTreeNumbaType, "_reduce_args", jit_options={"cache": True})
def _ol_reduce_args(self):
    dtype = self.field_dict['data'].dtype
    if dtype != nb.types.float32:
        dtype = nb.types.float64

    # functions to retrieve the parameters of the tree
    func_leafsize = ckdtree_ct.leafsize[dtype]
    func_size = ckdtree_ct.size[dtype]
    func_nodesize = ckdtree_ct.nodesize[dtype]
    func_copy_tree = ckdtree_ct.copy_tree[dtype]

    def _reduce_args_impl(self):
        leafsize = func_leafsize(self.ckdtree)
        size_bytes = func_size(self.ckdtree) * func_nodesize(self.ckdtree)
        # copy the tree into a fresh buffer
        tree_buffer = np.empty(size_bytes, dtype=np.uint8)
        size_copied = func_copy_tree(self.ckdtree, tree_buffer.ctypes)
        if size_copied != size_bytes:
            raise ValueError("__getstate__ failed.")
        return (tree_buffer, self.data, self.root_bbox, leafsize, self.idx)
    
    return _reduce_args_impl


@overload_method(KDTreeNumbaType, "build_index", jit_options={"nogil": True, "cache" : True, "fastmath": True})
def _ol_build_index(self, leafsize, balanced=False, compact=False):
    # choose the appropriate methods based on the data type
    dtype = self.field_dict['data'].dtype
    if dtype != nb.types.float32:
        dtype = nb.types.float64

    func_init = ckdtree_ct.init[dtype]
    func_build = ckdtree_ct.build[dtype]
    func_free = ckdtree_ct.free[dtype]

    def _build_index_impl(self, leafsize, balanced=False, compact=False):
        n_data, n_features = self.data.shape
        if self.ckdtree != 0:
           func_free(self.ckdtree)
        self.ckdtree = func_init(0, 0, self.data.ctypes, self.idx.ctypes, n_data, n_features, leafsize, self.root_bbox[0].ctypes, self.root_bbox[1].ctypes)
        compact_ = 1 if compact else 0
        balanced_ = 1 if balanced else 0
        func_build(self.ckdtree, 0, n_data, self.root_bbox[0].ctypes, self.root_bbox[1].ctypes, balanced_, compact_)

    return _build_index_impl


@overload_method(KDTreeNumbaType, "query", jit_options={"nogil": True, "cache": True, "fastmath": True, 'parallel': False})
def _ol_query(self, X, k=1, p=2, eps=0.0, distance_upper_bound=np.inf):
    # choose the appropriate methods based on the data type
    dtype = self.field_dict['data'].dtype
    if dtype == nb.types.float32:
        dtype_npy = np.float32
    else:
        dtype = nb.types.float64
        dtype_npy = np.float64

    func_query_knn = ckdtree_ct.query_knn[dtype] 
    
    # single threaded case
    def _query_impl(self, X, k=1, p=2, eps=0.0, distance_upper_bound=np.inf):
        n_features = self.data.shape[1]
        xx = _convert_to_valid_input(X, n_features, dtype_npy)
        n_queries = xx.shape[0]
        if p < 1:
            raise ValueError("Only p-norms with 1<=p<=infinity permitted")

        dd = np.empty((n_queries, k), dtype=dtype_npy)
        ii = np.full((n_queries, k), fill_value=-1, dtype=INT_TYPE)
        nn = np.empty((n_queries,), dtype=INT_TYPE)

        for i in range(n_queries):
            func_query_knn(self.ckdtree, dd[i].ctypes, ii[i].ctypes, nn[i:i+1].ctypes,
                            xx[i].ctypes, 1, k, eps, p, distance_upper_bound)
        return dd, ii, nn

    return _query_impl
    

# parallel version
@overload_method(KDTreeNumbaType, "query_parallel", jit_options={"nogil": True, "cache": False, "fastmath": True, 'parallel': True})
def _ol_query_parallel(self, X, k=1, p=2, eps=0.0, distance_upper_bound=np.inf):
    # choose the appropriate methods based on the data type
    dtype = self.field_dict['data'].dtype
    if dtype == nb.types.float32:
        dtype_npy = np.float32
    else:
        dtype = nb.types.float64
        dtype_npy = np.float64

    func_query_knn = ckdtree_ct.query_knn[dtype]

    def _query_parallel_impl(self, X, k=1, p=2, eps=0.0, distance_upper_bound=np.inf):
        n_features = self.data.shape[1]
        xx = _convert_to_valid_input(X, n_features, dtype_npy)
        n_queries = xx.shape[0]

        if p < 1:
            raise ValueError("Only p-norms with 1<=p<=infinity permitted")
        
        dd = np.empty((n_queries, k), dtype=dtype_npy)
        ii = np.full((n_queries, k), fill_value=-1, dtype=INT_TYPE)
        nn = np.empty((n_queries,), dtype=INT_TYPE)

        for i in nb.prange(n_queries):
            func_query_knn(self.ckdtree, dd[i].ctypes, ii[i].ctypes, nn[i:i + 1].ctypes,
                            xx[i].ctypes, 1, k, eps, p, distance_upper_bound)
        return dd, ii, nn

    return _query_parallel_impl


@overload_method(KDTreeNumbaType, "query_radius", jit_options={"nogil": True, "cache": True, "fastmath": True, 'parallel': False})
def _ol_query_radius(self, X, r, p=2.0, eps=0.0, return_sorted=False, return_length=False):
    # choose the appropriate methods based on the data type
    dtype = self.field_dict['data'].dtype
    if dtype == nb.types.float32:
        dtype_npy = np.float32
    else:
        dtype = nb.types.float64
        dtype_npy = np.float64

    broadcast_r = (r != nb.types.Array)

    func_query_knn = ckdtree_ct.query_radius[dtype]
    radius_result_set_get_size = ckdtree_ct.radius_result_set_get_size
    radius_result_set_copy_and_free = ckdtree_ct.radius_result_set_copy_and_free

    result_array_type = types.int64[:]

    # noinspection PyShadowingNames
    def _query_radius_impl(self, X, r, p=2.0, eps=0.0, return_sorted=False, return_length=False):
        n_features = self.data.shape[1]
        xx = _convert_to_valid_input(X, n_features, dtype_npy)
        n_queries = xx.shape[0]

        # broadcast a scalar r into the appropriate shape
        if broadcast_r:
            r_ = np.broadcast_to(r, n_queries)
        else:
            r_ = _convert_to_valid_input/(r, 1, dtype_npy).squeeze()

            if r_.shape != (n_queries,):
                raise ValueError("Invalid shape for r. Must be broadcastable to the number of queries.")

        if p < 1:
            raise ValueError("Only p-norms with 1<=p<=infinity permitted")

        # prepare result list
        results_list = nb.typed.List.empty_list(item_type=result_array_type, allocated=n_queries)
        results_list.extend([np.empty(0, dtype=np.int64) for i in range(n_queries)])

        for i in range(n_queries):
            result_set = func_query_knn(self.ckdtree, xx[i].ctypes, 1, r_[i], eps, p, return_length, return_sorted)
            # copy the result set into a separate buffer owned by python
            num_results = radius_result_set_get_size(result_set)
            results = np.empty(num_results, dtype=np.int64)
            radius_result_set_copy_and_free(result_set, results.ctypes)
            results_list[i] = results

        return results_list

    return _query_radius_impl


@overload_method(KDTreeNumbaType, "query_radius_parallel", jit_options={"nogil": True, "cache": False, "fastmath": True, 'parallel': True})
def _ol_query_radius_parallel(self, X, r, p=2.0, eps=0.0, return_sorted=False, return_length=False):
    # choose the appropriate methods based on the data type
    dtype = self.field_dict['data'].dtype
    if dtype == nb.types.float32:
        dtype_npy = np.float32
    else:
        dtype = nb.types.float64
        dtype_npy = np.float64

    broadcast_r = (r != nb.types.Array)

    func_query_knn = ckdtree_ct.query_radius[dtype]
    radius_result_set_get_size = ckdtree_ct.radius_result_set_get_size
    radius_result_set_copy_and_free = ckdtree_ct.radius_result_set_copy_and_free

    result_array_type = types.int64[:]

    # noinspection PyShadowingNames
    def _query_radius_parallel_impl(self, X, r, p=2.0, eps=0.0, return_sorted=False, return_length=False):
        n_features = self.data.shape[1]
        xx = _convert_to_valid_input(X, n_features, dtype_npy)
        n_queries = xx.shape[0]

        # broadcast a scalar r into the appropriate shape
        if broadcast_r:
            r_ = np.broadcast_to(r, n_queries)
        else:
            r_ = _convert_to_valid_input/(r, 1, dtype_npy).squeeze()
            if r_.shape != (n_queries,):
                raise ValueError("Invalid shape for r. Must be broadcastable to the number of queries.")
            
        if p < 1:
            raise ValueError("Only p-norms with 1<=p<=infinity permitted")

        # prepare result list
        results_list = nb.typed.List.empty_list(item_type=result_array_type, allocated=n_queries)
        results_list.extend([np.empty(0, dtype=np.int64) for i in range(n_queries)])

        for i in nb.prange(n_queries):
            result_set = func_query_knn(self.ckdtree, xx[i].ctypes, 1, r_[i], eps, p, return_length, return_sorted)
            # copy the result set into a separate buffer owned by python
            num_results = radius_result_set_get_size(result_set)
            results = np.empty(num_results, dtype=np.int64)
            radius_result_set_copy_and_free(result_set, results.ctypes)
            results_list[i] = results

        return results_list

    return _query_radius_parallel_impl


@nb.njit(nogil=True, cache=True)
def _make_kdtree(data, root_bbox, idx, leafsize=10, balanced=False, compact=False) -> KDTreeNumbaType:
    # create the transparent underlying c object by calling the function appropriate to the data dtype
    ckdtree = np.uint64(0)  # leave the c object empty for now
    kdtree = KDTreeType(ckdtree, root_bbox, data, idx)
    kdtree.build_index(leafsize, balanced, compact)
    return kdtree


def _restore_kdtree_impl(tree_buffer, data, root_bbox, leafsize, indices):
    # this is a stub for numba overload
    pass


@nb.extending.overload(_restore_kdtree_impl, jit_options={'nogil': True, 'fastmath': True, "cache": True})
def _ol_restore_kdtree_impl(tree_buffer, data, root_bbox, leafsize, indices):
    dtype = data.dtype
    if dtype == nb.types.float32:
        dtype_npy = np.float32
    else:
        dtype = nb.types.float64
        dtype_npy = np.float64

    func_init = ckdtree_ct.init[dtype]

    def _restore_kdtree_impl_impl(tree_buffer, data, root_bbox, leafsize, indices):
        data_conv = data.astype(dtype_npy) # is this really needed?
        n_data, n_features = data.shape
        ckdtree = np.uint64(0)  # leave the c object empty for now
        kdtree = KDTreeType(ckdtree, root_bbox, data_conv, indices)
        # call init with the existing tree
        kdtree.ckdtree = func_init(tree_buffer.ctypes, tree_buffer.size, data_conv.ctypes, indices.ctypes, n_data, n_features, leafsize, root_bbox[0].ctypes, root_bbox[1].ctypes)
        return kdtree

    return _restore_kdtree_impl_impl


# wrapper function to call the overloaded function above
@nb.njit(cache=True)
def _restore_kdtree(tree_buffer, data, root_bbox, leafsize, indices):
    return _restore_kdtree_impl(tree_buffer, data, root_bbox, leafsize, indices)


# python constructor function
def KDTree(data: DataArray, leafsize: int = 10, compact: bool = False, balanced: bool = False, root_bbox: Optional[DataArray] = None) -> KDTreeType:
    """
    Represents a KDTree usable from python as well as within numba functions. The tree is represented by the custom structref KDTreeType and 
    can be freely constructed and passed between numba and regular python code. 
    The underlying implementation uses a modified version of the ckdtree 
    available in the scipy package (https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.cKDTree.html) 
    and is comparable regarding performance and results.
    The data array stored by reference if possible, so no additional copy will be made. The user has to ensure, 
    that the data is not modified during the lifetime of the tree to avoid corruption.

    The resulting KDTree is fully pickable and can be passed between processes and stored on disk. The data array will be serialized with the tree in that case.

    Args:
        data: The data to build the tree from as an array of shape (n, n_features). This array is not copied into the tree but stored by reference so it must 
              not be modified during the lifetime of the tree.
        leafsize: The maximum number of points in the leafs of the tree. Larger values result in a smaller tree but higher search time. Defaults to 10.
        compact: If True, the KDTree is built to shrink the hyperrectangles to the actual data range. 
                 This usually gives a more compact tree that is robust against degenerated input data and 
                 gives faster queries at the expense of longer build time. Defaults to False.
        balanced: If True, the median is used to split the hyperrectangles instead of the midpoint. 
                  This usually gives a more compact tree and faster queries at the expense of longer build time.
                  Defaults to False.
        root_bbox: Initial bounding box to use for the root. If None (default), the bounds will be determined from the data array before building the tree.

    Returns:
        The python proxy objects containing the KDTree.
    """
    if data.dtype == np.float32:
        conv_dtype = np.float32
    else:
        conv_dtype = np.float64

    data = np.ascontiguousarray(data).astype(conv_dtype)
    n_data, n_features = data.shape

    if root_bbox is None:
        # compute the bounding box
        mins = np.amin(data, axis=0) if n_data > 0 else np.zeros(n_features, dtype=conv_dtype)
        maxes = np.amax(data, axis=0) if n_data > 0 else np.zeros(n_features, dtype=conv_dtype)
        root_bbox = np.vstack((mins, maxes))

    root_bbox = np.ascontiguousarray(root_bbox, dtype=conv_dtype)

    idx = np.arange(n_data, dtype=INT_TYPE)

    kdtree = _make_kdtree(data, root_bbox, idx, leafsize, balanced, compact)
    return kdtree


# numba constructor implementation
@nb.extending.overload(KDTree, jit_options={'nogil': True, 'fastmath': True, 'cache': True})
def KDTree_numba(data, leafsize: int = 10, compact: bool = False, balanced: bool = False, root_bbox=None):
    if data.dtype == nb.types.float32:
        conv_dtype = nb.types.float32
        finfo = np.finfo(np.float32)

    else:
        conv_dtype = nb.types.float64
        finfo = np.finfo(np.float64)

    cmax = finfo.max
    cmin = finfo.min

    def KDTree_impl(data, leafsize=10, compact=False, balanced=False, root_bbox=None):
        data = np.ascontiguousarray(data).astype(conv_dtype)
        n_data, n_features = data.shape

        if root_bbox is None:
            # compute the bounding box
            root_bbox_ = np.empty((2, 3), dtype=data.dtype)
            root_bbox_[0] = cmax
            root_bbox_[1] = cmin

            for i in range(data.shape[0]):
                for j in range(data.shape[1]):
                    if data[i, j] < root_bbox_[0, j]:
                        root_bbox_[0, j] = data[i, j]
                    if data[i, j] > root_bbox_[1, j]:
                        root_bbox_[1, j] = data[i, j]
        else:
            root_bbox_ = root_bbox
        root_bbox__ = np.ascontiguousarray(root_bbox_).astype(conv_dtype)

        idx = np.arange(n_data, dtype=INT_TYPE)

        kdtree = _make_kdtree(data, root_bbox__, idx, leafsize, balanced, compact)

        return kdtree

    return KDTree_impl



import numpy as np
from numba_kdtree import KDTree


data = np.random.random(3_000_000).reshape(-1, 3).astype(np.float32)
kdtree = KDTree(data, leafsize=10)

# query the nearest neighbors of the first 100 points
distances, indices, num_neighbors = kdtree.query(data[:100], k=30)

# query all points in a radius around the first 100 points
indices = kdtree.query_radius(data[:100], r=0.5, return_sorted=True)


import requests
import tarfile
from pathlib import Path
import sys, os
from typing import Union
import numpy as np
from numpy.typing import NDArray, DTypeLike
import plyfile

__all__ = ['load_standford_bunny']


def load_standford_bunny(directory: Union[Path, os.PathLike, str, bytes] = "test_data",
                         dtype: DTypeLike = np.float32,
                         chunk_size: int = 8192) -> NDArray:
    """
    Downloads the stanford bunny into the specified directory and returns the PointCloud

    Args:
        directory: The directory to load.
    Returns:
        A numpy array containing the stanford bunny position data.
    """
    directory = Path(directory)
    directory.mkdir(exist_ok=True)
    bunny_path = directory / "bunny.tar.gz"

    # download the file if it does not exist
    if not bunny_path.exists():
        url = 'http://graphics.stanford.edu/pub/3Dscanrep/bunny.tar.gz'
        with open(bunny_path, 'wb') as f:

            response = requests.get(url, stream=True)
            total_length = response.headers.get('content-length')

            if total_length is None:  # no content length header
                f.write(response.content)
            else:
                for data in response.iter_content(chunk_size=chunk_size):
                    f.write(data)

    # Extract the data
    bunny_tar_file = tarfile.open(bunny_path)

    data_dir = bunny_path.parent / "bunny_data"
    data_dir.mkdir(exist_ok=True)
    bunny_tar_file.extractall(data_dir)
    bunny_tar_file.close()

    # Load the ply file from the data path

    with open(data_dir / "bunny" / "reconstruction" / "bun_zipper.ply", 'rb') as f:
        plydata = plyfile.PlyData.read(f)
        x = plydata.elements[0].data['x']
        y = plydata.elements[0].data['y']
        z = plydata.elements[0].data['z']
        positions = np.stack((x, y, z), axis=1)

    positions = positions.astype(dtype, copy=False)

    return positions



from setuptools import setup, find_packages, Extension
from distutils.command.build_ext import build_ext as build_ext_orig
from distutils.sysconfig import get_python_inc
from os.path import join, dirname
from distutils.sysconfig import customize_compiler
import platform

# standalone import of a module (https://stackoverflow.com/a/58423785)
def import_module_from_path(path):
    """Import a module from the given path without executing any code above it
    """
    import importlib
    import pathlib
    import sys

    module_path = pathlib.Path(path).resolve()
    module_name = module_path.stem  # 'path/x.py' -> 'x'
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    module = importlib.util.module_from_spec(spec)

    if module not in sys.modules:
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
    else:
        module = sys.modules
    return module

inc_dirs = [get_python_inc()]
inc_dirs.append(join(dirname(dirname(__file__)), 'src', 'ckdtree'))

ckdtree_src = ['init.cpp',
               'build.cpp',
               'query.cpp',
               'query_radius.cpp']

ckdtree_src = [join('src', 'ckdtree', x) for x in ckdtree_src]


class CTypesExtension(Extension): pass


class build_ext(build_ext_orig):
    def build_extension(self, ext):
        self._ctypes = isinstance(ext, CTypesExtension)
        if self._ctypes:
            customize_compiler(self.compiler)
            try:
                self.compiler.compiler_so.remove("-Wstrict-prototypes")
            except (AttributeError, ValueError):
                pass
        return super().build_extension(ext)

    def get_export_symbols(self, ext):
        if self._ctypes:
            return ext.export_symbols
        return super().get_export_symbols(ext)

    def get_ext_filename(self, ext_name):
        if self._ctypes:
            if platform.system() == "Windows":
                return ext_name + '.lib'
            else:
                return ext_name + '.so'
        return super().get_ext_filename(ext_name)


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

if platform.system() == "Windows":
    extra_compile_args = ['/O2', '/DKDTREE_COMPILING=1']
else:
    extra_compile_args = ['-fPIC', '-shared', '-O3', '-DKDTREE_COMPILING=1']
    if platform.system() == "Darwin":
        extra_compile_args.append('-std=c++11')

module = CTypesExtension('numba_kdtree._ckdtree',
                   sources=ckdtree_src,
                   include_dirs=inc_dirs,
                   extra_compile_args=extra_compile_args)

setup(
    name="numba-kdtree",
    author="Felix Igelbrink",
    author_email="felix.igelbrink@uni-osnabrueck.de",
    description="A kdtree implementation for numba.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mortacious/numba-kdtree",
    project_urls={
        "Bug Tracker": "https://github.com/mortacious/numba-kdtree/issues",
    },
    packages=find_packages(exclude=['examples', 'tests']),
    ext_modules=[module],
    install_requires=[
        'numba>=0.52',
    ],
    extras_require={
        'tests': ['pytest', 'scipy', 'requests', 'plyfile'],
    },
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development",
    ],
    python_requires=">=3.7",
    cmdclass={'build_ext': build_ext},
    zip_safe=True,
    setup_requires=['setuptools_scm'],
    use_scm_version = {
        "write_to": "numba_kdtree/_version.py",
        "version_scheme": "release-branch-semver"
    },
)
